#include "task.h"
#include "paging.h"
#include "kheap.h"

// The currently running task.
volatile task_t* current_task;

// The start of the task linked list.
volatile task_t* ready_queue;

// Some externs are needed to access members in paging.c...
extern page_directory_t* kernel_directory;
extern page_directory_t* current_directory;
extern tss_entry_t tss_entry;

extern ULONG read_eip();

// The next available process ID.
ULONG next_pid = 1;

void tasking_install()
{
    cli(); // disable interrupts

    // Relocate the stack so we know where it is.
    move_stack((void*)0x00190000, 0x10000);

    printformat("after moving stack, ESP: %x \n", fetchESP());

    // Initialise the first task (kernel task)
    current_task = ready_queue = (task_t*)k_malloc(sizeof(task_t),0,0);

    printformat("first task structure: %x ", current_task);

    current_task->id = next_pid++;
    current_task->esp = current_task->ebp = 0;
    current_task->eip = 0;
    current_task->page_directory = current_directory;
    current_task->next = 0;
    current_task->kernel_stack = k_malloc(KERNEL_STACK_SIZE,1,0)+KERNEL_STACK_SIZE;

    printformat("kernel_stack: %x\n", current_task->kernel_stack);

    sti(); // enable interrupts
}

void move_stack(void* new_stack_start, ULONG size)
{
  ULONG i;
  // Allocate some space for the new stack.
  for( i=(ULONG)new_stack_start; i>=((ULONG)new_stack_start-size); i-=0x1000)
  {
    // General-purpose stack is in user-mode.
    alloc_frame( get_page(i, 1, current_directory), 0 /* User mode */, 1 /* Is writable */ );
  }

  // Flush the TLB by reading and writing the page directory address again.
  ULONG pd_addr;
  asm volatile("mov %%cr3, %0" :  "=r" (pd_addr));
  asm volatile("mov %0, %%cr3" : : "r" (pd_addr));

  // Old ESP and EBP, read from registers.
  ULONG old_stack_pointer; asm volatile("mov %%esp, %0" : "=r" (old_stack_pointer));
  ULONG old_base_pointer;  asm volatile("mov %%ebp, %0" : "=r" (old_base_pointer));

  // Offset to add to old stack addresses to get a new stack address.
  ULONG offset            = (ULONG)new_stack_start - initial_esp;

  // New ESP and EBP.
  ULONG new_stack_pointer = old_stack_pointer + offset;
  ULONG new_base_pointer  = old_base_pointer  + offset;

  // Copy the stack.
  k_memcpy((void*)new_stack_pointer, (void*)old_stack_pointer, initial_esp-old_stack_pointer);

  // Backtrace through the original stack, copying new values into the new stack.
  for(i = (ULONG)new_stack_start; i > (ULONG)new_stack_start-size; i -= 4)
  {
    ULONG tmp = * (ULONG*)i;
    // If the value of tmp is inside the range of the old stack, assume it is a base pointer and remap it.
    // This will unfortunately remap ANY value in this range, whether they are base pointers or not.
    if (( old_stack_pointer < tmp) && (tmp < initial_esp))
    {
      tmp = tmp + offset;
      ULONG *tmp2 = (ULONG*)i;
      *tmp2 = tmp;
    }
  }

  // Change stacks.
  asm volatile("mov %0, %%esp" : : "r" (new_stack_pointer));
  asm volatile("mov %0, %%ebp" : : "r" (new_base_pointer));
}

void task_switch()
{
    // If we haven't initialised tasking yet, just return.
    if(!current_task) return;

    cli();
    // Read esp, ebp now for saving later on.
    ULONG esp, ebp, eip;

    eip = read_eip();
    if(eip == 0x12345) return; // Have we just switched tasks?

    asm volatile("mov %%esp, %0" : "=r"(esp));
    asm volatile("mov %%ebp, %0" : "=r"(ebp));

    //old_task
    current_task->eip = eip;
    current_task->esp = esp;
    current_task->ebp = ebp;

    //--------------------------
    //old_task ==> new_task
    current_task = current_task->next;
    if(!current_task) current_task = ready_queue;

    // new_task
    current_directory = current_task->page_directory;
    set_kernel_stack(current_task->kernel_stack+KERNEL_STACK_SIZE);
    tss_entry.eip = eip = current_task->eip; //TEST
    tss_entry.cr3 = current_directory->physicalAddr; //TEST
    esp = current_task->esp;
    ebp = current_task->ebp;
    //TSS_log(&tss_entry); //TEST

    asm volatile("       \
    cli;                 \
    mov %%ebx, %%esp;    \
    mov %%edx, %%ebp;    \
    mov %%eax, %%cr3;    \
    mov $0x12345, %%eax; \
    sti;                 \
    jmp *%%ecx;          "
               : : "c"(eip), "b"(esp), "d"(ebp), "a"(current_directory->physicalAddr):"%edi","%esi" );
}

int fork()
{
    // We are modifying kernel structures, and so cannot be interrupted.
    cli();

    // Take a pointer to this process' task struct for later reference.
    task_t* parent_task = (task_t*)current_task;

    // Clone the address space.
    page_directory_t* directory = clone_directory(current_directory);

    // Create a new process.
    task_t* new_task = (task_t*)k_malloc(sizeof(task_t),0,0);

    new_task->id  = next_pid++;
    new_task->esp = new_task->ebp = 0;
    new_task->eip = 0;
    new_task->page_directory = directory;
    new_task->kernel_stack = k_malloc(KERNEL_STACK_SIZE,1,0);
    new_task->next = 0;

    // Add it to the end of the ready queue. Find the end of the ready queue ...
    task_t* tmp_task = (task_t*)ready_queue;
    while (tmp_task->next)
        tmp_task = tmp_task->next;
    tmp_task->next = new_task; // ... and extend it

    ULONG eip = read_eip(); // This will be the entry point for the new process

    // We could be the parent or the child here - check.
    if (current_task == parent_task)
    {
        // We are the parent, so set up the esp/ebp/eip for our child
        ULONG esp; asm volatile("mov %%esp, %0" : "=r"(esp));
        ULONG ebp; asm volatile("mov %%ebp, %0" : "=r"(ebp));
        new_task->esp = esp;
        new_task->ebp = ebp;
        new_task->eip = eip;

        sti();               // All finished: Reenable interrupts.
        return new_task->id; // And by convention return the PID of the child
    }
    else
        return 0; // We are the child - by convention return 0
}

int getpid() { return current_task->id; }

void switch_to_user_mode()
{
    // Set up our kernel stack.
    set_kernel_stack(current_task->kernel_stack+KERNEL_STACK_SIZE); //tss_entry.esp0 = ...;

    printformat("After set_kernel_stack ==> tss_entry.esp0: %x\n",current_task->kernel_stack+KERNEL_STACK_SIZE);

    // Set up a stack structure for switching to user mode (ring 3 => RPL = 0x3)
    // 0x200 is IF flag (enabling interrupts)
    // 0x20|0x3=0x23  // 0x18|0x3=0x1B
    // $1F means "the address of the next label '1:', searching forward"

/*
    //User Mode:
      asm volatile("  \
      cli; \
      mov $0x23, %ax; \
      mov %ax, %ds; \
      mov %ax, %es; \
      mov %ax, %fs; \
      mov %ax, %gs; \
                    \
      mov %esp, %eax; \
      pushl $0x23; \
      pushl %esp; \
      pushf; \
      pop %eax; \
      or $0x200, %eax; \
      push %eax; \
      pushl $0x1B; \
      push $1f; \
      iret; \
    1: \
      ");
*/
    //Kernel Mode:
    asm volatile("  \
      cli; \
      mov $0x20, %ax; \
      mov %ax, %ds; \
      mov %ax, %es; \
      mov %ax, %fs; \
      mov %ax, %gs; \
                    \
      mov %esp, %eax; \
      pushl $0x20; \
      pushl %esp; \
      pushf; \
      pop %eax; \
      or $0x200, %eax; \
      push %eax; \
      pushl $0x18; \
      push $1f; \
      iret; \
    1: \
      ");
}

void task_log(task_t* t)
{
    settextcolor(5,0);
    printformat("id: %d ", t->id);               // Process ID
    printformat("ebp: %x ",t->ebp);              // Base pointer
    printformat("esp: %x ",t->esp);              // Stack pointer
    printformat("eip: %x ",t->eip);              // Instruction pointer
    printformat("PD: %x ", t->page_directory);   // Page directory.
    printformat("k_stack: %x ",t->kernel_stack); // Kernel stack location.
    printformat("next: %x\n",  t->next);         // The next task in a linked list.
}

void TSS_log(tss_entry_t* tss)
{
    settextcolor(6,0);
    printformat("prev_tss: %x ", tss->prev_tss);
    printformat("esp0: %x ", tss->esp0);
    printformat("ss0: %x ", tss->ss0);
    printformat("esp1: %x ", tss->esp1);
    printformat("ss1: %x ", tss->ss1);
    printformat("esp2: %x ", tss->esp2);
    printformat("ss2: %x ", tss->ss2);
    printformat("cr3: %x ", tss->cr3);
    printformat("eip: %x ", tss->eip);
    printformat("eflags: %x ", tss->eflags);
    printformat("eax: %x ", tss->eax);
    printformat("ecx: %x ", tss->ecx);
    printformat("edx: %x ", tss->edx);
    printformat("ebx: %x ", tss->ebx);
    printformat("esp: %x ", tss->esp);
    printformat("ebp: %x ", tss->ebp);
    printformat("esi: %x ", tss->esi);
    printformat("edi: %x ", tss->edi);
    printformat("es: %x ", tss->es);
    printformat("cs: %x ", tss->cs);
    printformat("ss: %x ", tss->ss);
    printformat("ds: %x ", tss->ds);
    printformat("fs: %x ", tss->fs);
    printformat("gs: %x ", tss->gs);
    printformat("ldt: %x ", tss->ldt);
    printformat("trap: %x ", tss->trap); //only 0 or 1
    printformat("iomap_base: %x ", tss->iomap_base);
}

ULONG task_switch1(ULONG esp)
{
    // If we haven't initialised tasking yet, just return.
    if(!current_task) return esp;

    current_task->esp = esp;   // save esp

    //old_task ==> new_task
    current_task = current_task->next;
    if(!current_task) current_task = ready_queue;

    // new_task
    current_directory = current_task->page_directory;
    set_kernel_stack(current_task->kernel_stack+KERNEL_STACK_SIZE);
    tss_entry.cr3 = current_directory->physicalAddr;

    return current_task->esp;  // return esp
}
